-- LibGroup
-- @class file
-- @name LibGroup-1.0.lua
-- @release 1.0

--[[
   Copyright 2009 Christo Kotze (onaforeignshore)

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
]]

assert(LibStub:GetLibrary("CallbackHandler-1.0", true), "LibGroup-1.0 requires CallbackHandler-1.0")

local MAJOR_VERSION = "LibGroup-1.0"
local MINOR_VERSION = tonumber(("$Rev: 2 $"):match("%d+")) or 0

local lib, oldMinor = LibStub:NewLibrary(MAJOR_VERSION, MINOR_VERSION)
if not lib then return end

lib.callbacks = lib.callbacks or LibStub:GetLibrary("CallbackHandler-1.0"):New(lib)
local callbacks = lib.callbacks

-- localise global functions and variables
local GetNumRaidMembers = _G.GetNumRaidMembers
local GetRaidRosterInfo = _G.GetRaidRosterInfo
local GetNumPartyMembers = _G.GetNumPartyMembers
local UnitName = _G.UnitName
local UnitIsPartyLeader = _G.UnitIsPartyLeader
local UnitLevel = _G.UnitLevel
local UnitClass = _G.UnitClass
local UnitIsConnected = _G.UnitIsConnected
local UnitIsDeadOrGhost = _G.UnitIsDeadOrGhost
local new, del
do
	-- unused tables go in this set
	-- if the garbage collector comes around, they'll be collected properly
	local cache = setmetatable({}, {__mode='k'})
	
	--- Return a table
	-- @usage local t = new()
	-- @return a blank table
	function new()
		local t = next(cache)
		if t then
			cache[t] = nil
			return t
		end
		
		return {}
	end
	
	--- Delete a table, clearing it and putting it back into the queue
	-- @usage local t = new()
	-- t = del(t)
	-- @return nil
	function del(t)
		wipe(t)
		cache[t] = true
		return nil
	end
end

-- create a frame for the library
local frame
if lib.frame then
	frame = lib.frame
	frame:UnregisterAllEvents()
	frame:SetScript("OnEvent", nil)
	frame:SetScript("OnUpdate", nil)
else
	frame = CreateFrame("Frame", MAJOR_VERSION .. "-Frame")
	lib.frame = frame
end
frame:SetScript("OnEvent", function(this, event, ...)
	this[event](lib, ...)
end)
frame:RegisterEvent("RAID_ROSTER_UPDATE")
frame:RegisterEvent("PARTY_MEMBERS_CHANGED")

local inGroup = false
local group = {}
local groupType = 0
local masterLooter = 0

function frame:RAID_ROSTER_UPDATE()
	if GetNumRaidMembers() >= 1 then
		for i = 1, GetNumRaidMembers() do
			local name, rank, subgroup, level, _, fileName, _, online, dead, _, isML = GetRaidRosterInfo(i)
			if (not group[name]) and inGroup then
				callbacks:Fire("Joined", name)
			end
			if dead == "" then dead = nil end
			if name then
				group[name] = group[name] or {}
				group[name].name = name
				group[name].rank = rank
				group[name].level = level
				group[name].subgroup = subgroup
				group[name].class = fileName
				if group[name].online == 1 and not online then
					callbacks:Fire("Disconnected", name)
				end
				if group[name].online == -1 and online then
					callbacks:Fire("Connected", name)
				end
				group[name].online = online or -1 -- trickery to get above callbacks to work
				group[name].dead = dead
				group[name].id = "raid"..i
				group[name].updated = true
				if isML then
					masterLooter = name
				end
			end
		end
		if not inGroup then
			inGroup = true
			groupType = "RAID"
			callbacks:Fire("Joined", UnitName("player"))
		end
		for i, v in pairs(group) do
			if not v.updated then
				group[i] = nil
				callbacks:Fire("Left", i)
			else
				v.updated = nil
			end
		end
	else
		inGroup = false
		groupType = nil
		callbacks:Fire("Left", UnitName("player"))
		-- clear the table for next time
		for i, _ in pairs(group) do
			group[i] = nil
		end
	end
end

function frame:PARTY_MEMBERS_CHANGED()
	if GetNumRaidMembers() > 0 then return end
	if GetNumPartyMembers() >= 1 then
		for i = 0, GetNumPartyMembers() do
			local id
			if (i == 0) then
				id = "player"
			else
				id = "party"..i
			end
			local name, server = UnitName(id)
			if server and server ~= ""  then
				name = name.."-"..server
			end
			local rank = UnitIsPartyLeader(id)
			local level = UnitLevel(id)
			local _, fileName = UnitClass(id)
			local online = UnitIsConnected(id)
			local dead = UnitIsDeadOrGhost(id)
			if (not group[name]) and inGroup then
				callbacks:Fire("Joined", name)
			end
			if dead == "" then dead = nil end
			if name then
				group[name] = group[name] or {}
				group[name].name = name
				if rank then
					group[name].rank = 2
				else
					group[name].rank = 0
				end
				group[name].level = level
				group[name].class = fileName
				if group[name].online == 1 and not online then
					callbacks:Fire("Disconnected", name)
				end
				if group[name].online == -1 and online then
					callbacks:Fire("Connected", name)
				end
				group[name].online = online or -1 -- trickery to get above callbacks to work
				group[name].dead = dead
				group[name].id = id
				group[name].updated = true
			end
		end
		if not inGroup then
			inGroup = true
			groupType = "PARTY"
			callbacks:Fire("Joined", UnitName("player"))
		end
		for i, v in pairs(group) do
			if not v.updated then
				group[i] = nil
				callbacks:Fire("Left", i)
			else
				v.updated = nil
			end
		end
	else
		inGroup = false
		groupType = nil
		callbacks:Fire("Left", UnitName("player"))
		-- clear the table for next time
		for i, _ in pairs(group) do
			group[i] = nil
		end
	end
end

function lib:IsInGroup()
	return inGroup
end

function lib:GetClass(name)
	name = name or UnitName("player")
	return (group[name] and group[name].class) or "UNKNOWN"
end

function lib:GetClassColor(name)
	name = name or UnitName("player")
	local class = group[name] and group[name].class
	if not class then return 0.8, 0.8, 0.8 end
	local c = RAID_CLASS_COLORS[class]
	return c.r, c.g, c.b
end

function lib:GetClassHexColor(name)
	name = name or UnitName("player")
	local class = group[name] and group[name].class
	if not class then return 'cccccc' end
	local c = RAID_CLASS_COLORS[class]
	return ("%02X%02X%02X"):format(c.r * 255, c.g * 255, c.b * 255)
end

function lib:GetGroupType()
	return groupType
end

function lib:GetLevel(name)
	name = name or UnitName("player")
	return (group[name] and group[name].level) or 0
end

function lib:GetRank(name)
	name = name or UnitName("player")
	return (group[name] and group[name].rank) or 0
end

function lib:GetSubgroup(name)
	name = name or UnitName("player")
	return (group[name] and group[name].subgroup) or 0
end

function lib:GetUnitId(name)
	name = name or UnitName("player")
	return (group[name] and group[name].id) or "none"
end

function lib:IsDead(name)
	name = name or UnitName("player")
	return (group[name] and group[name].dead) or 0
end

function lib:IsMasterLooter(name)
	name = name or UnitName("player")
	return name == masterLooter or 0
end

function lib:IsOnline(name)
	name = name or UnitName("player")
	return (group[name] and group[name].online) or 0
end

function lib:GetPlayers()
	return players
end

function lib:GetPlayersWithClassColor()
	local t = {}
	for k, _ in pairs(group) do
		t[k] = "|cff"..lib:GetClassHexColor(k)..k.."|r"
	end
	return t
end

local sorts; sorts = {
	NAME = function(a, b)
		return group[a].name < group[b].name
	end,
	LEVEL = function(a, b)
		if group[a].level == group[b].level then
			return sorts.NAME(a, b)
		else
			return group[a].level < group[b].level
		end
	end,
	CLASS = function(a, b)
		if group[a].class == group[b].class then
			return sorts.NAME(a, b)
		else
			return group[a].class < group[b].class
		end
	end,
	RANK = function(a, b)
		if group[a].rank == group[b].rank then
			return sorts.NAME(a, b)
		else
			return group[a].rank < group[b].rank
		end
	end,
	ONLINE = function(a, b)
		if group[a].online == group[b].online then
			return sorts.NAME(a, b)
		else
			return group[a] -- 1 or 0 (nil), you see.
		end
	end,
}

local iter = function(t)
	local n = t.n + 1
	t.n = n
	return t[n] or del(t)
end

function lib:Iterate(includeOffline, sort)
	local sortFunc = sorts[sort or "NAME"]
	if not sortFunc then
		error('Argument #2 must be "NAME", "LEVEL", "CLASS", "RANK", "ONLINE", or nil.')
	end
	
	local tmp = new()
	for _, v in pairs(group) do
		if includeOffline then
			tmp[#tmp + 1] = v.name
		else
			if v.online then tmp[#tmp + 1] = v.name end
		end
	end
	
	table.sort(tmp, sortFunc)
	tmp.n = 0
	
	return iter, tmp, nil
end

if inGroup then
	frame:Show()
else
	frame:Hide()
end